import os
from typing import Any, Dict, List, Optional, Tuple, Union

from pydantic import PrivateAttr

from redisvl.utils.rerank.base import BaseReranker


class VoyageAIReranker(BaseReranker):
    """
    The VoyageAIReranker class uses VoyageAI's API to rerank documents based on an
    input query.

    This reranker is designed to interact with VoyageAI's /rerank API,
    requiring an API key for authentication. The key can be provided
    directly in the `api_config` dictionary or through the `VOYAGE_API_KEY`
    environment variable. User must obtain an API key from VoyageAI's website
    (https://dash.voyageai.com/). Additionally, the `voyageai` python
    client must be installed with `pip install voyageai`.

    .. code-block:: python

        from redisvl.utils.rerank import VoyageAIReranker

        # set up the VoyageAI reranker with some configuration
        reranker = VoyageAIReranker(rank_by=["content"], limit=2)
        # rerank raw search results based on user input/query
        results = reranker.rank(
            query="your input query text here",
            docs=[
                {"content": "document 1"},
                {"content": "document 2"},
                {"content": "document 3"}
            ]
        )
    """

    _client: Any = PrivateAttr()
    _aclient: Any = PrivateAttr()

    def __init__(
        self,
        model: str,
        rank_by: Optional[List[str]] = None,
        limit: int = 5,
        return_score: bool = True,
        api_config: Optional[Dict] = None,
        **kwargs,
    ):
        """
        Initialize the VoyageAIReranker with specified model, ranking criteria,
        and API configuration.

        Parameters:
            model (str): The identifier for the VoyageAI model used for reranking.
            rank_by (Optional[List[str]]): Optional list of keys specifying the
                attributes in the documents that should be considered for
                ranking. None means ranking will rely on the model's default
                behavior.
            limit (int): The maximum number of results to return after
                reranking. Must be a positive integer.
            return_score (bool): Whether to return scores alongside the
                reranked results.
            api_config (Optional[Dict], optional): Dictionary containing the API key.
                Defaults to None.

        Raises:
            ImportError: If the voyageai library is not installed.
            ValueError: If the API key is not provided.
        """
        super().__init__(
            model=model, rank_by=rank_by, limit=limit, return_score=return_score
        )
        self._initialize_clients(api_config, **kwargs)

    def _initialize_clients(self, api_config: Optional[Dict], **kwargs):
        """
        Setup the VoyageAI clients using the provided API key or an
        environment variable.
        """
        # Dynamic import of the voyageai module
        try:
            from voyageai import AsyncClient, Client
        except ImportError:
            raise ImportError(
                "VoyageAI reranker requires the voyageai library. \
                    Please install with `pip install voyageai`"
            )

        # Fetch the API key from api_config or environment variable
        api_key = (
            api_config.get("api_key") if api_config else os.getenv("VOYAGE_API_KEY")
        )
        if not api_key:
            raise ValueError(
                "VoyageAI API key is required. "
                "Provide it in api_config or set the VOYAGE_API_KEY environment variable."
            )
        self._client = Client(api_key=api_key, **kwargs)
        self._aclient = AsyncClient(api_key=api_key, **kwargs)

    def _preprocess(
        self, query: str, docs: Union[List[Dict[str, Any]], List[str]], **kwargs
    ):
        """
        Prepare and validate reranking config based on provided input and
        optional overrides.
        """
        limit = kwargs.get("limit", self.limit)
        return_score = kwargs.get("return_score", self.return_score)
        truncation = kwargs.get("truncation")

        if not isinstance(docs, list):
            raise ValueError("Docs to rerank should either be a list of str or dict.")

        if all(isinstance(doc, dict) and "content" in doc for doc in docs):
            texts = [
                str(doc["content"])
                for doc in docs
                if isinstance(doc, dict) and "content" in doc
            ]
        else:
            texts = [str(doc) for doc in docs]

        reranker_kwargs = {
            "query": query,
            "documents": texts,
            "model": self.model,
            "top_k": limit,
        }
        if truncation is not None:
            reranker_kwargs["truncation"] = truncation

        return reranker_kwargs, return_score

    @staticmethod
    def _postprocess(
        docs: Union[List[Dict[str, Any]], List[str]],
        rankings: List[Any],
    ) -> Tuple[List[Any], List[float]]:
        """
        Post-process the initial list of documents to include ranking scores,
        if specified.
        """
        reranked_docs, scores = [], []
        for item in rankings.results:  # type: ignore
            scores.append(item.relevance_score)
            reranked_docs.append(docs[item.index])
        return reranked_docs, scores

    def rank(
        self, query: str, docs: Union[List[Dict[str, Any]], List[str]], **kwargs
    ) -> Union[Tuple[List[Dict[str, Any]], List[float]], List[Dict[str, Any]]]:
        """
        Rerank documents based on the provided query using the VoyageAI rerank API.

        This method processes the user's query and the provided documents to
        rerank them in a manner that is potentially more relevant to the
        query's context.

        Parameters:
            query (str): The user's search query.
            docs (Union[List[Dict[str, Any]], List[str]]): The list of documents
                to be ranked, either as dictionaries or strings.

        Returns:
            Union[Tuple[Union[List[Dict[str, Any]], List[str]], float], List[Dict[str, Any]]]: The reranked list of documents and optionally associated scores.
        """
        reranker_kwargs, return_score = self._preprocess(query, docs, **kwargs)
        rankings = self._client.rerank(**reranker_kwargs)
        reranked_docs, scores = self._postprocess(docs, rankings)
        if return_score:
            return reranked_docs, scores
        return reranked_docs

    async def arank(
        self, query: str, docs: Union[List[Dict[str, Any]], List[str]], **kwargs
    ) -> Union[Tuple[List[Dict[str, Any]], List[float]], List[Dict[str, Any]]]:
        """
        Rerank documents based on the provided query using the VoyageAI rerank API.

        This method processes the user's query and the provided documents to
        rerank them in a manner that is potentially more relevant to the
        query's context.

        Parameters:
            query (str): The user's search query.
            docs (Union[List[Dict[str, Any]], List[str]]): The list of documents
                to be ranked, either as dictionaries or strings.

        Returns:
            Union[Tuple[Union[List[Dict[str, Any]], List[str]], float], List[Dict[str, Any]]]: The reranked list of documents and optionally associated scores.
        """
        reranker_kwargs, return_score = self._preprocess(query, docs, **kwargs)
        rankings = await self._aclient.rerank(**reranker_kwargs)
        reranked_docs, scores = self._postprocess(docs, rankings)
        if return_score:
            return reranked_docs, scores
        return reranked_docs



from redisvl.query.aggregate import (
    AggregateHybridQuery,
    AggregationQuery,
    MultiVectorQuery,
    Vector,
)
from redisvl.query.hybrid import HybridQuery
from redisvl.query.query import (
    BaseQuery,
    BaseVectorQuery,
    CountQuery,
    FilterQuery,
    RangeQuery,
    TextQuery,
    VectorQuery,
    VectorRangeQuery,
)

__all__ = [
    "BaseQuery",
    "BaseVectorQuery",
    "VectorQuery",
    "FilterQuery",
    "RangeQuery",
    "VectorRangeQuery",
    "CountQuery",
    "TextQuery",
    "AggregationQuery",
    "AggregateHybridQuery",
    "MultiVectorQuery",
    "Vector",
]



from typing import Any, Dict, List, Literal, Optional, Set, Union

from redis.commands.search.query import Filter

from redisvl.query.filter import FilterExpression
from redisvl.redis.utils import array_to_buffer
from redisvl.utils.full_text_query_helper import FullTextQueryHelper

_IMPORT_ERROR_MESSAGE = "Hybrid queries require Redis >= 8.4.0 and redis-py>=7.1.0"


class HybridQuery:
    """
    A hybrid search query that combines text search and vector similarity, with configurable fusion methods.

    .. code-block:: python

        from redisvl.query import HybridQuery
        from redisvl.index import SearchIndex

        index = SearchIndex.from_yaml("path/to/index.yaml")

        query = HybridQuery(
            text="example text",
            text_field_name="text_field",
            vector=[0.1, 0.2, 0.3],
            vector_field_name="vector_field",
            text_scorer="BM25STD",
            yield_text_score_as="text_score",
            yield_vsim_score_as="vector_similarity",
            combination_method="LINEAR",
            linear_alpha=0.3,
            yield_combined_score_as="hybrid_score",
            num_results=10,
            return_fields=["field1", "field2"],
            stopwords="english",
        )

        results = index.query(query)

    See Also:
        - `FT.HYBRID command documentation <https://redis.io/docs/latest/commands/ft.hybrid>`_
        - `redis-py hybrid_search documentation <https://redis.readthedocs.io/en/stable/redismodules.html#redis.commands.search.commands.SearchCommands.hybrid_search>`_
    """

    def __init__(
        self,
        text: str,
        text_field_name: str,
        vector: Union[bytes, List[float]],
        vector_field_name: str,
        vector_param_name: str = "vector",
        text_scorer: str = "BM25STD",
        yield_text_score_as: Optional[str] = None,
        vector_search_method: Optional[Literal["KNN", "RANGE"]] = None,
        knn_ef_runtime: int = 10,
        range_radius: Optional[float] = None,
        range_epsilon: float = 0.01,
        yield_vsim_score_as: Optional[str] = None,
        filter_expression: Optional[Union[str, FilterExpression]] = None,
        combination_method: Optional[Literal["RRF", "LINEAR"]] = None,
        rrf_window: int = 20,
        rrf_constant: int = 60,
        linear_alpha: float = 0.3,
        yield_combined_score_as: Optional[str] = None,
        dtype: str = "float32",
        num_results: Optional[int] = 10,
        return_fields: Optional[List[str]] = None,
        stopwords: Optional[Union[str, Set[str]]] = "english",
        text_weights: Optional[Dict[str, float]] = None,
    ):
        """
        Instantiates a HybridQuery object.

        Args:
            text: The text to search for.
            text_field_name: The text field name to search in.
            vector: The vector to perform vector similarity search.
            vector_field_name: The vector field name to search in.
            vector_param_name: The name of the parameter substitution containing the vector blob.
            text_scorer: The text scorer to use. Options are {TFIDF, TFIDF.DOCNORM,
                BM25STD, BM25STD.NORM, BM25STD.TANH, DISMAX, DOCSCORE, HAMMING}. Defaults to "BM25STD". For more
                information about supported scoring algorithms,
                see https://redis.io/docs/latest/develop/ai/search-and-query/advanced-concepts/scoring/
            yield_text_score_as: The name of the field to yield the text score as.
            vector_search_method: The vector search method to use. Options are {KNN, RANGE}. Defaults to None.
            knn_ef_runtime: The exploration factor parameter for HNSW, optional if `vector_search_method` is "KNN".
            range_radius: The search radius to use, required if `vector_search_method` is "RANGE".
            range_epsilon: The epsilon value to use, optional if `vector_search_method` is "RANGE"; defines the
                accuracy of the search.
            yield_vsim_score_as: The name of the field to yield the vector similarity score as.
            filter_expression: The filter expression to use for both the text and vector searches. Defaults to None.
            combination_method: The combination method to use. Options are {RRF, LINEAR}. If not specified, the server
                defaults to RRF. If "RRF" is specified, then at least one of `rrf_window` or `rrf_constant` must be
                provided. If "LINEAR" is specified, then at least one of `linear_alpha` or `linear_beta` must be
                provided.
            rrf_window: The window size to use for the reciprocal rank fusion (RRF) combination method. Limits
                fusion scope.
            rrf_constant: The constant to use for the reciprocal rank fusion (RRF) combination method. Controls decay
                of rank influence.
            linear_alpha: The weight of the text query for the linear combination method (LINEAR).
            yield_combined_score_as: The name of the field to yield the combined score as.
            dtype: The data type of the vector. Defaults to "float32".
            num_results: The number of results to return.
            return_fields: The fields to return. Defaults to None.
            stopwords (Optional[Union[str, Set[str]]], optional): The stopwords to remove from the
                provided text prior to search-use. If a string such as "english" "german" is
                provided then a default set of stopwords for that language will be used. if a list,
                set, or tuple of strings is provided then those will be used as stopwords.
                Defaults to "english". if set to "None" then no stopwords will be removed.

                Note: This parameter controls query-time stopword filtering (client-side).
                For index-level stopwords configuration (server-side), see IndexInfo.stopwords.
                Using query-time stopwords with index-level STOPWORDS 0 is counterproductive.
            text_weights (Optional[Dict[str, float]]): The importance weighting of individual words
                within the query text. Defaults to None, as no modifications will be made to the
                text_scorer score.

        Raises:
            ImportError: If redis-py>=7.1.0 is not installed.
            TypeError: If the stopwords are not a set, list, or tuple of strings.
            ValueError: If the text string is empty, or if the text string becomes empty after
                stopwords are removed.
            ValueError: If `vector_search_method` is defined and isn't one of {KNN, RANGE}.
            ValueError: If `vector_search_method` is "KNN" and `knn_k` is not provided.
            ValueError: If `vector_search_method` is "RANGE" and `range_radius` is not provided.
        """
        try:
            from redis.commands.search.hybrid_query import (
                CombineResultsMethod,
                HybridPostProcessingConfig,
            )
        except (ImportError, ModuleNotFoundError):
            raise ImportError(_IMPORT_ERROR_MESSAGE)

        self.postprocessing_config = HybridPostProcessingConfig()
        if num_results:
            self.postprocessing_config.limit(offset=0, num=num_results)
        if return_fields:
            self.postprocessing_config.load(*(f"@{f}" for f in return_fields))

        self._ft_helper = FullTextQueryHelper(
            stopwords=stopwords,
            text_weights=text_weights,
        )

        query_string = self._ft_helper.build_query_string(
            text, text_field_name, filter_expression
        )

        if isinstance(vector, bytes):
            vector_data = vector
        else:
            vector_data = array_to_buffer(vector, dtype)

        self.params = {
            vector_param_name: vector_data,
        }

        self.query = build_base_query(
            text_query=query_string,
            vector_param_name=vector_param_name,
            vector_field_name=vector_field_name,
            text_scorer=text_scorer,
            yield_text_score_as=yield_text_score_as,
            vector_search_method=vector_search_method,
            num_results=num_results,
            knn_ef_runtime=knn_ef_runtime,
            range_radius=range_radius,
            range_epsilon=range_epsilon,
            yield_vsim_score_as=yield_vsim_score_as,
            filter_expression=filter_expression,
        )

        if combination_method:
            self.combination_method: Optional[CombineResultsMethod] = (
                build_combination_method(
                    combination_method=combination_method,
                    rrf_window=rrf_window,
                    rrf_constant=rrf_constant,
                    linear_alpha=linear_alpha,
                    yield_score_as=yield_combined_score_as,
                )
            )
        else:
            self.combination_method = None


def build_base_query(
    text_query: str,
    vector_param_name: str,
    vector_field_name: str,
    text_scorer: str = "BM25STD",
    yield_text_score_as: Optional[str] = None,
    vector_search_method: Optional[Literal["KNN", "RANGE"]] = None,
    num_results: Optional[int] = None,
    knn_ef_runtime: Optional[int] = None,
    range_radius: Optional[float] = None,
    range_epsilon: Optional[float] = None,
    yield_vsim_score_as: Optional[str] = None,
    filter_expression: Optional[Union[str, FilterExpression]] = None,
):
    """Build a Redis HybridQuery for performing hybrid search.

    Args:
        text_query: The query for the text search.
        vector_param_name: The name of the parameter substitution containing the vector blob.
        vector_field_name: The vector field name to search in.
        text_scorer: The text scorer to use. Options are {TFIDF, TFIDF.DOCNORM,
            BM25STD, BM25STD.NORM, BM25STD.TANH, DISMAX, DOCSCORE, HAMMING}. Defaults to "BM25STD". For more
            information about supported scroring algorithms,
            see https://redis.io/docs/latest/develop/ai/search-and-query/advanced-concepts/scoring/
        yield_text_score_as: The name of the field to yield the text score as.
        vector_search_method: The vector search method to use. Options are {KNN, RANGE}. Defaults to None.
        num_results: The number of nearest neighbors to return, required if `vector_search_method` is "KNN".
        knn_ef_runtime: The exploration factor parameter for HNSW, optional if `vector_search_method` is "KNN".
        range_radius: The search radius to use, required if `vector_search_method` is "RANGE".
        range_epsilon: The epsilon value to use, optional if `vector_search_method` is "RANGE"; defines the
            accuracy of the search.
        yield_vsim_score_as: The name of the field to yield the vector similarity score as.
        filter_expression: The filter expression to use for the vector similarity search. Defaults to None.

    Notes:
        If RRF combination method is used, then at least one of `rrf_window` or `rrf_constant` must be provided.
        If LINEAR combination method is used, then at least one of `linear_alpha` or `linear_beta` must be provided.

    Raises:
        ImportError: If redis-py>=7.1.0 is not installed.
        ValueError: If `vector_search_method` is defined and isn't one of {KNN, RANGE}.
        ValueError: If `vector_search_method` is "KNN" and `knn_k` is not provided.
        ValueError: If `vector_search_method` is "RANGE" and `range_radius` is not provided.

    Returns:
        A Redis HybridQuery object that defines the text and vector searches to be performed.
    """
    try:
        from redis.commands.search.hybrid_query import HybridQuery as RedisHybridQuery
        from redis.commands.search.hybrid_query import (
            HybridSearchQuery,
            HybridVsimQuery,
            VectorSearchMethods,
        )
    except (ImportError, ModuleNotFoundError):
        raise ImportError(_IMPORT_ERROR_MESSAGE)

    # Serialize the full-text search query
    search_query = HybridSearchQuery(
        query_string=text_query,
        scorer=text_scorer,
        yield_score_as=yield_text_score_as,
    )

    # Serialize vector similarity search method and params, if specified
    vsim_search_method: Optional[VectorSearchMethods] = None
    vsim_search_method_params: Dict[str, Any] = {}
    if vector_search_method == "KNN":
        vsim_search_method = VectorSearchMethods.KNN
        if not num_results:
            raise ValueError(
                "Must provide `num_results` if vector_search_method is KNN"
            )

        vsim_search_method_params["K"] = num_results
        if knn_ef_runtime:
            vsim_search_method_params["EF_RUNTIME"] = knn_ef_runtime

    elif vector_search_method == "RANGE":
        vsim_search_method = VectorSearchMethods.RANGE
        if not range_radius:
            raise ValueError("Must provide RADIUS if vector_search_method is RANGE")

        vsim_search_method_params["RADIUS"] = range_radius
        if range_epsilon:
            vsim_search_method_params["EPSILON"] = range_epsilon

    elif vector_search_method is not None:
        raise ValueError(f"Unknown vector search method: {vector_search_method}")

    if isinstance(filter_expression, FilterExpression):
        filter_expression = str(filter_expression)

    if filter_expression and filter_expression != "*":
        vsim_filter = Filter("FILTER", str(filter_expression))
    else:
        vsim_filter = None

    # Serialize the vector similarity query
    vsim_query = HybridVsimQuery(
        vector_field_name="@" + vector_field_name,
        vector_data="$" + vector_param_name,
        vsim_search_method=vsim_search_method,
        vsim_search_method_params=vsim_search_method_params,
        filter=vsim_filter,
        yield_score_as=yield_vsim_score_as,
    )

    return RedisHybridQuery(
        search_query=search_query,
        vector_similarity_query=vsim_query,
    )


def build_combination_method(
    combination_method: Literal["RRF", "LINEAR"],
    rrf_window: Optional[int] = None,
    rrf_constant: Optional[float] = None,
    linear_alpha: Optional[float] = None,
    yield_score_as: Optional[str] = None,
):
    """Build a configuration for combining hybrid search scores.

    Args:
        combination_method: The combination method to use. Options are {RRF, LINEAR}.
        rrf_window: The window size to use for the reciprocal rank fusion (RRF) combination method. Limits
            fusion scope.
        rrf_constant: The constant to use for the reciprocal rank fusion (RRF) combination method. Controls decay
            of rank influence.
        linear_alpha: The weight of the first query for the linear combination method (LINEAR).
        yield_score_as: The name of the field to yield the combined score as.

    Raises:
        ImportError: If redis-py>=7.1.0 is not installed.
        ValueError: If `combination_method` is defined and isn't one of {RRF, LINEAR}.
        ValueError: If `combination_method` is "RRF" and neither `rrf_window` nor `rrf_constant` is provided.
        ValueError: If `combination_method` is "LINEAR" and neither `linear_alpha` nor `linear_beta` is provided.

    Returns:
        A CombineResultsMethod object that defines how the text and vector scores should be combined.
    """
    try:
        from redis.commands.search.hybrid_query import (
            CombinationMethods,
            CombineResultsMethod,
        )
    except (ImportError, ModuleNotFoundError):
        raise ImportError(_IMPORT_ERROR_MESSAGE)

    method_params: Dict[str, Any] = {}
    if combination_method == "RRF":
        method = CombinationMethods.RRF
        if rrf_window:
            method_params["WINDOW"] = rrf_window
        if rrf_constant:
            method_params["CONSTANT"] = rrf_constant

    elif combination_method == "LINEAR":
        method = CombinationMethods.LINEAR
        if linear_alpha:
            method_params["ALPHA"] = linear_alpha
            method_params["BETA"] = 1 - linear_alpha

    else:
        raise ValueError(f"Unknown combination method: {combination_method}")

    if yield_score_as:
        method_params["YIELD_SCORE_AS"] = yield_score_as

    if not method_params:
        raise ValueError(
            "No parameters provided for combination method - must provide at least one parameter."
        )

    return CombineResultsMethod(
        method=method,
        **method_params,
    )



import pytest

from redisvl.utils.rerank.hf_cross_encoder import HFCrossEncoderReranker


@pytest.fixture(scope="session")
def reranker():
    return HFCrossEncoderReranker()


def test_rank_documents(reranker):
    docs = ["document one", "document two", "document three"]
    query = "search query"

    reranked_docs, scores = reranker.rank(query, docs)

    assert isinstance(reranked_docs, list)
    assert len(reranked_docs) == reranker.limit
    assert all(isinstance(score, float) for score in scores)


@pytest.mark.asyncio
async def test_async_rank_documents(reranker):
    docs = ["document one", "document two", "document three"]
    query = "search query"

    reranked_docs, scores = await reranker.arank(query, docs)

    assert isinstance(reranked_docs, list)
    assert len(reranked_docs) == reranker.limit
    assert all(isinstance(score, float) for score in scores)


def test_bad_input(reranker):
    with pytest.raises(ValueError):
        reranker.rank("", [])  # Empty query

    with pytest.raises(TypeError):
        reranker.rank(123, ["valid document"])  # Invalid type for query

    with pytest.raises(TypeError):
        reranker.rank("valid query", "not a list")  # Invalid type for documents


def test_rerank_empty(reranker):
    docs = []
    query = "search query"

    reranked_docs = reranker.rank(query, docs, return_score=False)

    assert isinstance(reranked_docs, list)
    assert len(reranked_docs) == 0



import re
from typing import Any


class TypeInferrer:
    """Infers the type of a field based on its value."""

    GEO_PATTERN = re.compile(
        r"^\s*[-+]?([1-8]?\d(\.\d+)?|90(\.0+)?),\s*[-+]?(180(\.0+)?|((1[0-7]\d)|([1-9]?\d))(\.\d+)?)\s*$"
    )

    TYPE_METHOD_MAP = {
        "numeric": "_is_numeric",
        "geo": "_is_geographic",
        "tag": "_is_tag",
        "text": "_is_text",
    }

    @classmethod
    def infer(cls, value: Any) -> str:
        """Infers the field type for a given value.

        Args:
            value: The value to infer the type of.

        Returns:
            The inferred field type as a string.

        Raises:
            ValueError: If the type cannot be inferred.
        """
        for type_name, method_name in cls.TYPE_METHOD_MAP.items():
            if getattr(cls, method_name)(value):
                return type_name
        raise ValueError(f"Unable to infer type for value: {value}")

    @classmethod
    def _is_numeric(cls, value: Any) -> bool:
        """Check if the value is numeric."""
        if not isinstance(value, (int, float, str)):
            return False
        try:
            float(value)
            return True
        except (ValueError, TypeError):
            return False

    @classmethod
    def _is_tag(cls, value: Any) -> bool:
        """Check if the value is a tag."""
        return isinstance(value, (list, set, tuple)) and all(
            isinstance(v, str) for v in value
        )

    @classmethod
    def _is_text(cls, value: Any) -> bool:
        """Check if the value is text."""
        return isinstance(value, str)

    @classmethod
    def _is_geographic(cls, value: Any) -> bool:
        """Check if the value is a geographic coordinate."""
        return isinstance(value, str) and cls.GEO_PATTERN.match(value) is not None



import os
import warnings

import numpy as np
import pytest

from redisvl.extensions.cache.embeddings.embeddings import EmbeddingsCache
from redisvl.utils.utils import create_ulid
from redisvl.utils.vectorize import (
    AzureOpenAITextVectorizer,
    BedrockVectorizer,
    CohereTextVectorizer,
    CustomVectorizer,
    HFTextVectorizer,
    MistralAITextVectorizer,
    OpenAITextVectorizer,
    VertexAIVectorizer,
    VoyageAIVectorizer,
)

# Constants for testing
TEST_TEXT = "This is a test sentence."
TEST_TEXTS = ["This is the first test sentence.", "This is the second test sentence."]
TEST_VECTOR = [1.1, 2.2, 3.3, 4.4]


@pytest.fixture
def embeddings_cache(client):
    """Create a real EmbeddingsCache for testing with a unique namespace."""
    # Use a unique prefix for this test run to avoid conflicts
    unique_prefix = f"test_cache_{create_ulid()}"

    # Create the cache with a short TTL
    cache = EmbeddingsCache(name=unique_prefix, ttl=10, redis_client=client)

    yield cache

    cache.clear()


@pytest.fixture(
    params=[
        HFTextVectorizer,
        OpenAITextVectorizer,
        VertexAIVectorizer,
        CohereTextVectorizer,
        AzureOpenAITextVectorizer,
        BedrockVectorizer,
        MistralAITextVectorizer,
        CustomVectorizer,
        VoyageAIVectorizer,
    ]
)
def vectorizer(request):
    if request.param == HFTextVectorizer:
        return request.param()
    elif request.param == OpenAITextVectorizer:
        return request.param()
    elif request.param == VertexAIVectorizer:
        return request.param()
    elif request.param == CohereTextVectorizer:
        return request.param()
    elif request.param == MistralAITextVectorizer:
        return request.param()
    elif request.param == VoyageAIVectorizer:
        return request.param(model="voyage-large-2")
    elif request.param == AzureOpenAITextVectorizer:
        return request.param(
            model=os.getenv("AZURE_OPENAI_DEPLOYMENT_NAME", "text-embedding-ada-002")
        )
    elif request.param == BedrockVectorizer:
        return request.param(
            model=os.getenv("BEDROCK_MODEL_ID", "amazon.titan-embed-text-v2:0")
        )
    elif request.param == CustomVectorizer:

        def embed(content):
            return TEST_VECTOR

        def embed_many(contents):
            return [TEST_VECTOR] * len(contents)

        async def aembed_func(content):
            return TEST_VECTOR

        async def aembed_many_func(contents):
            return [TEST_VECTOR] * len(contents)

        return request.param(embed=embed, embed_many=embed_many)


@pytest.fixture
def cached_vectorizer(embeddings_cache):
    """Create a simple custom vectorizer for testing."""

    def embed(content):
        return TEST_VECTOR

    def embed_many(contents):
        return [TEST_VECTOR] * len(contents)

    async def aembed(content):
        return TEST_VECTOR

    async def aembed_many(contents):
        return [TEST_VECTOR] * len(contents)

    return CustomVectorizer(
        embed=embed,
        embed_many=embed_many,
        aembed=aembed,
        aembed_many=aembed_many,
        cache=embeddings_cache,
    )


@pytest.fixture
def custom_embed_func():
    def embed(content: str):
        return TEST_VECTOR

    return embed


@pytest.fixture
def custom_embed_class():
    class MyEmbedder:
        def embed(self, content: str):
            return TEST_VECTOR

        def embed_with_args(self, content: str, max_len=None):
            return TEST_VECTOR[0:max_len]

        def embed_many(self, contents):
            return [[1.1, 2.2, 3.3], [4.4, 5.5, 6.6]]

        def embed_many_with_args(self, contents, param=True):
            if param:
                return [[1.0, 2.0, 3.0], [4.0, 5.0, 6.0]]
            else:
                return [[6.0, 5.0, 4.0], [3.0, 2.0, 1.0]]

    return MyEmbedder


@pytest.mark.requires_api_keys
def test_vectorizer_embed(vectorizer):
    text = TEST_TEXT
    if isinstance(vectorizer, CohereTextVectorizer):
        embedding = vectorizer.embed(text, input_type="search_document")
    elif isinstance(vectorizer, VoyageAIVectorizer):
        embedding = vectorizer.embed(text, input_type="document")
    else:
        embedding = vectorizer.embed(text)

    assert isinstance(embedding, list)
    assert len(embedding) == vectorizer.dims


@pytest.mark.requires_api_keys
def test_vectorizer_embed_many(vectorizer):
    texts = TEST_TEXTS
    if isinstance(vectorizer, CohereTextVectorizer):
        embeddings = vectorizer.embed_many(texts, input_type="search_document")
    elif isinstance(vectorizer, VoyageAIVectorizer):
        embeddings = vectorizer.embed_many(texts, input_type="document")
    else:
        embeddings = vectorizer.embed_many(texts)

    assert isinstance(embeddings, list)
    assert len(embeddings) == len(texts)
    assert all(
        isinstance(emb, list) and len(emb) == vectorizer.dims for emb in embeddings
    )


def test_vectorizer_with_cache(cached_vectorizer):
    """Test the complete cache flow - miss, store, hit."""
    # First call - should be a cache miss
    first_result = cached_vectorizer.embed(TEST_TEXT)
    assert first_result == TEST_VECTOR

    # Second call - should be a cache hit
    second_result = cached_vectorizer.embed(TEST_TEXT)
    assert second_result == TEST_VECTOR

    # Verify it's actually using the cache by checking the cached value exists
    cached_entry = cached_vectorizer.cache.get(
        content=TEST_TEXT, model_name=cached_vectorizer.model
    )
    assert cached_entry is not None
    assert cached_entry["embedding"] == TEST_VECTOR


def test_vectorizer_with_cache_skip(cached_vectorizer):
    """Test embedding with skip_cache=True."""
    # Store a value in the cache
    cached_vectorizer.embed(TEST_TEXT)

    # Call embed with skip_cache=True - should bypass cache
    cached_vectorizer.cache.drop(content=TEST_TEXT, model_name=cached_vectorizer.model)

    # Store a deliberately different value in the cache
    cached_vectorizer.cache.set(
        content=TEST_TEXT,
        model_name=cached_vectorizer.model,
        embedding=[9.9, 8.8, 7.7, 6.6],
    )

    # Now call with skip_cache=True
    result = cached_vectorizer.embed(TEST_TEXT, skip_cache=True)

    # Should generate fresh result, not use cached value
    assert result == TEST_VECTOR

    # Cache should still have the original value
    cached_entry = cached_vectorizer.cache.get(
        content=TEST_TEXT, model_name=cached_vectorizer.model
    )
    assert cached_entry["embedding"] == [9.9, 8.8, 7.7, 6.6]


def test_vectorizer_with_cache_many(cached_vectorizer):
    """Test embedding many texts with partial cache hits/misses."""
    # Store an embedding for the first text only
    cached_vectorizer.cache.set(
        content=TEST_TEXTS[0],
        model_name=cached_vectorizer.model,
        embedding=[0.1, 0.2, 0.3, 0.4],
    )

    # Call embed_many - should hit cache for first text, miss for second
    results = cached_vectorizer.embed_many(TEST_TEXTS)

    # Verify results
    assert results[0] == [0.1, 0.2, 0.3, 0.4]  # From cache
    assert results[1] == TEST_VECTOR  # Generated

    # Both should now be in cache
    for text in TEST_TEXTS:
        assert cached_vectorizer.cache.exists(
            content=text, model_name=cached_vectorizer.model
        )


def test_vectorizer_with_cached_metadata(cached_vectorizer):
    """Test passing metadata through to the cache."""
    # Call embed with metadata
    test_metadata = {"source": "test", "importance": "high"}
    cached_vectorizer.embed(TEST_TEXT, metadata=test_metadata)

    # Verify metadata was stored in cache
    cached_entry = cached_vectorizer.cache.get(
        content=TEST_TEXT, model_name=cached_vectorizer.model
    )
    assert cached_entry["metadata"] == test_metadata


@pytest.mark.asyncio
async def test_vectorizer_with_cache_async(cached_vectorizer):
    """Test async embedding with cache."""
    # First call - should be a cache miss
    first_result = await cached_vectorizer.aembed(TEST_TEXT)
    assert first_result == TEST_VECTOR

    # Second call - should be a cache hit
    second_result = await cached_vectorizer.aembed(TEST_TEXT)
    assert second_result == TEST_VECTOR

    # Verify it's actually using the cache
    cached_entry = await cached_vectorizer.cache.aget(
        content=TEST_TEXT, model_name=cached_vectorizer.model
    )
    assert cached_entry is not None
    assert cached_entry["embedding"] == TEST_VECTOR


@pytest.mark.asyncio
async def test_vectorizer_with_cache_async_many(cached_vectorizer):
    """Test async embedding many texts with partial cache hits/misses."""
    # Store an embedding for the first text only
    await cached_vectorizer.cache.aset(
        content=TEST_TEXTS[0],
        model_name=cached_vectorizer.model,
        embedding=[0.1, 0.2, 0.3, 0.4],
    )

    # Call aembed_many - should hit cache for first text, miss for second
    results = await cached_vectorizer.aembed_many(TEST_TEXTS)

    # Verify results
    assert results[0] == [0.1, 0.2, 0.3, 0.4]  # From cache
    assert results[1] == TEST_VECTOR  # Generated

    # Both should now be in cache
    for text in TEST_TEXTS:
        assert await cached_vectorizer.cache.aexists(
            content=text, model_name=cached_vectorizer.model
        )


@pytest.mark.requires_api_keys
def test_bedrock_bad_credentials():
    with pytest.raises(ValueError):
        BedrockVectorizer(
            api_config={
                "aws_access_key_id": "invalid",
                "aws_secret_access_key": "invalid",
            }
        )


@pytest.mark.requires_api_keys
def test_bedrock_invalid_model():
    with pytest.raises(ValueError):
        bedrock = BedrockVectorizer(model="invalid-model")
        bedrock.embed("test")


def test_custom_vectorizer_embed(custom_embed_class, custom_embed_func):
    custom_wrapper = CustomVectorizer(embed=custom_embed_func)
    embedding = custom_wrapper.embed("This is a test sentence.")
    assert embedding == TEST_VECTOR

    custom_wrapper = CustomVectorizer(embed=custom_embed_class().embed)
    embedding = custom_wrapper.embed("This is a test sentence.")
    assert embedding == TEST_VECTOR

    custom_wrapper = CustomVectorizer(embed=custom_embed_class().embed_with_args)
    embedding = custom_wrapper.embed("This is a test sentence.", max_len=4)
    assert embedding == TEST_VECTOR
    embedding = custom_wrapper.embed("This is a test sentence.", max_len=2)
    assert embedding == [1.1, 2.2]

    with pytest.raises(ValueError):
        invalid_vectorizer = CustomVectorizer(embed="hello")

    with pytest.raises(ValueError):
        invalid_vectorizer = CustomVectorizer(embed=42)

    with pytest.raises(ValueError):
        invalid_vectorizer = CustomVectorizer(embed={"foo": "bar"})

    def bad_arg_type(value: int):
        return [value]

    with pytest.raises(ValueError):
        invalid_vectorizer = CustomVectorizer(embed=bad_arg_type)

    def bad_return_type(text: str) -> str:
        return text

    with pytest.raises(ValueError):
        invalid_vectorizer = CustomVectorizer(embed=bad_return_type)


def test_custom_vectorizer_embed_many(custom_embed_class, custom_embed_func):
    custom_wrapper = CustomVectorizer(
        custom_embed_func, embed_many=custom_embed_class().embed_many
    )
    embeddings = custom_wrapper.embed_many(["test one.", "test two"])
    assert embeddings == [[1.1, 2.2, 3.3], [4.4, 5.5, 6.6]]

    custom_wrapper = CustomVectorizer(
        custom_embed_func, embed_many=custom_embed_class().embed_many
    )
    embeddings = custom_wrapper.embed_many(["test one.", "test two"])
    assert embeddings == [[1.1, 2.2, 3.3], [4.4, 5.5, 6.6]]

    custom_wrapper = CustomVectorizer(
        custom_embed_func, embed_many=custom_embed_class().embed_many_with_args
    )
    embeddings = custom_wrapper.embed_many(["test one.", "test two"], param=True)
    assert embeddings == [[1.0, 2.0, 3.0], [4.0, 5.0, 6.0]]
    embeddings = custom_wrapper.embed_many(["test one.", "test two"], param=False)
    assert embeddings == [[6.0, 5.0, 4.0], [3.0, 2.0, 1.0]]

    with pytest.raises(ValueError):
        invalid_vectorizer = CustomVectorizer(custom_embed_func, embed_many="hello")

    with pytest.raises(ValueError):
        invalid_vectorizer = CustomVectorizer(custom_embed_func, embed_many=42)

    with pytest.raises(ValueError):
        invalid_vectorizer = CustomVectorizer(
            custom_embed_func, embed_many={"foo": "bar"}
        )

    def bad_arg_type(value: int):
        return [value]

    with pytest.raises(ValueError):
        invalid_vectorizer = CustomVectorizer(
            custom_embed_func, embed_many=bad_arg_type
        )

    def bad_return_type(text: str) -> str:
        return text

    with pytest.raises(ValueError):
        invalid_vectorizer = CustomVectorizer(
            custom_embed_func, embed_many=bad_return_type
        )


@pytest.mark.requires_api_keys
@pytest.mark.parametrize(
    "vectorizer_",
    [
        AzureOpenAITextVectorizer,
        BedrockVectorizer,
        CohereTextVectorizer,
        CustomVectorizer,
        HFTextVectorizer,
        MistralAITextVectorizer,
        OpenAITextVectorizer,
        VertexAIVectorizer,
        VoyageAIVectorizer,
    ],
)
def test_default_dtype(vectorizer_):
    # test dtype defaults to float32
    if issubclass(vectorizer_, CustomVectorizer):
        vectorizer = vectorizer_(embed=lambda x, input_type=None: [1.0, 2.0, 3.0])
    elif issubclass(vectorizer_, AzureOpenAITextVectorizer):
        vectorizer = vectorizer_(
            model=os.getenv("AZURE_OPENAI_DEPLOYMENT_NAME", "text-embedding-ada-002")
        )
    else:
        vectorizer = vectorizer_()

    assert vectorizer.dtype == "float32"


@pytest.mark.requires_api_keys
@pytest.mark.parametrize(
    "vectorizer_",
    [
        AzureOpenAITextVectorizer,
        BedrockVectorizer,
        CohereTextVectorizer,
        CustomVectorizer,
        HFTextVectorizer,
        MistralAITextVectorizer,
        OpenAITextVectorizer,
        VertexAIVectorizer,
        VoyageAIVectorizer,
    ],
)
def test_vectorizer_dtype_assignment(vectorizer_):
    # test initializing dtype in constructor
    for dtype in ["float16", "float32", "float64", "bfloat16", "int8", "uint8"]:
        if issubclass(vectorizer_, CustomVectorizer):
            vectorizer = vectorizer_(embed=lambda x: [1.0, 2.0, 3.0], dtype=dtype)
        elif issubclass(vectorizer_, AzureOpenAITextVectorizer):
            vectorizer = vectorizer_(
                model=os.getenv(
                    "AZURE_OPENAI_DEPLOYMENT_NAME", "text-embedding-ada-002"
                ),
                dtype=dtype,
            )
        else:
            vectorizer = vectorizer_(dtype=dtype)

        assert vectorizer.dtype == dtype


@pytest.mark.requires_api_keys
@pytest.mark.parametrize(
    "vectorizer_",
    [
        AzureOpenAITextVectorizer,
        BedrockVectorizer,
        CohereTextVectorizer,
        HFTextVectorizer,
        MistralAITextVectorizer,
        OpenAITextVectorizer,
        VertexAIVectorizer,
        VoyageAIVectorizer,
    ],
)
def test_non_supported_dtypes(vectorizer_):
    with pytest.raises(ValueError):
        vectorizer_(dtype="float25")

    with pytest.raises(ValueError):
        vectorizer_(dtype=7)

    with pytest.raises(ValueError):
        vectorizer_(dtype=None)


@pytest.mark.requires_api_keys
@pytest.mark.asyncio
async def test_vectorizer_aembed(vectorizer):
    text = TEST_TEXT
    if isinstance(vectorizer, CohereTextVectorizer):
        embedding = await vectorizer.aembed(text, input_type="search_document")
    elif isinstance(vectorizer, VoyageAIVectorizer):
        embedding = await vectorizer.aembed(text, input_type="document")
    else:
        embedding = await vectorizer.aembed(text)
    assert isinstance(embedding, list)
    assert len(embedding) == vectorizer.dims


@pytest.mark.requires_api_keys
@pytest.mark.asyncio
async def test_vectorizer_aembed_many(vectorizer):
    texts = TEST_TEXTS
    if isinstance(vectorizer, CohereTextVectorizer):
        embeddings = await vectorizer.aembed_many(texts, input_type="search_document")
    elif isinstance(vectorizer, VoyageAIVectorizer):
        embeddings = await vectorizer.aembed_many(texts, input_type="document")
    else:
        embeddings = await vectorizer.aembed_many(texts)

    assert isinstance(embeddings, list)
    assert len(embeddings) == len(texts)
    assert all(
        isinstance(emb, list) and len(emb) == vectorizer.dims for emb in embeddings
    )


@pytest.mark.requires_api_keys
@pytest.mark.parametrize(
    "dtype,expected_type",
    [
        ("float32", float),  # Float dtype should return floats
        ("int8", int),  # Int8 dtype should return ints
        ("uint8", int),  # Uint8 dtype should return ints
    ],
)
def test_cohere_dtype_support(dtype, expected_type):
    """Test that CohereTextVectorizer properly handles different dtypes for embeddings."""
    text = TEST_TEXT
    texts = TEST_TEXTS

    # Create vectorizer with specified dtype
    vectorizer = CohereTextVectorizer(dtype=dtype)

    # Verify the correct mapping of dtype to Cohere embedding_types
    if dtype == "int8":
        assert vectorizer._get_cohere_embedding_type(dtype) == ["int8"]
    elif dtype == "uint8":
        assert vectorizer._get_cohere_embedding_type(dtype) == ["uint8"]
    else:
        # All other dtypes should map to float
        assert vectorizer._get_cohere_embedding_type(dtype) == ["float"]

    # Test single embedding
    embedding = vectorizer.embed(text, input_type="search_document")
    assert isinstance(embedding, list)
    assert len(embedding) == vectorizer.dims

    # Check that all elements are of the expected type
    assert all(
        isinstance(val, expected_type) for val in embedding
    ), f"Expected all elements to be {expected_type.__name__} for dtype {dtype}"

    # Test multiple embeddings
    embeddings = vectorizer.embed_many(texts, input_type="search_document")
    assert isinstance(embeddings, list)
    assert len(embeddings) == len(texts)
    assert all(
        isinstance(emb, list) and len(emb) == vectorizer.dims for emb in embeddings
    )

    # Check that all elements in all embeddings are of the expected type
    for emb in embeddings:
        assert all(
            isinstance(val, expected_type) for val in emb
        ), f"Expected all elements to be {expected_type.__name__} for dtype {dtype}"

    # Test as_buffer output format
    embedding_buffer = vectorizer.embed(
        text, input_type="search_document", as_buffer=True
    )
    assert isinstance(embedding_buffer, bytes)

    # Test embed_many with as_buffer=True
    buffer_embeddings = vectorizer.embed_many(
        texts, input_type="search_document", as_buffer=True
    )
    assert all(isinstance(emb, bytes) for emb in buffer_embeddings)

    # Compare dimensions between buffer and list formats
    assert len(np.frombuffer(embedding_buffer, dtype=dtype)) == len(embedding)


@pytest.mark.requires_api_keys
def test_cohere_embedding_types_warning():
    """Test that a warning is raised when embedding_types parameter is passed."""
    text = TEST_TEXT
    texts = TEST_TEXTS
    vectorizer = CohereTextVectorizer()

    # Test warning for single embedding
    with pytest.warns(UserWarning, match="embedding_types.*not supported"):
        embedding = vectorizer.embed(
            text,
            input_type="search_document",
            embedding_types=["uint8"],  # explicitly testing the anti-pattern here
        )
    assert isinstance(embedding, list)
    assert len(embedding) == vectorizer.dims

    # Test warning for multiple embeddings
    with pytest.warns(UserWarning, match="embedding_types.*not supported"):
        embeddings = vectorizer.embed_many(
            texts, input_type="search_document", embedding_types=["uint8"]
        )
    assert isinstance(embeddings, list)
    assert len(embeddings) == len(texts)


def test_deprecated_text_parameter_warning():
    """Test that using deprecated 'text' and 'texts' parameters emits deprecation warnings."""
    vectorizer = HFTextVectorizer(model="sentence-transformers/all-MiniLM-L6-v2")

    # Test single embed with deprecated 'text' parameter emits warning
    with pytest.warns(DeprecationWarning, match="Argument text is deprecated"):
        embedding = vectorizer.embed(text=TEST_TEXT)
    assert isinstance(embedding, list)
    assert len(embedding) == vectorizer.dims

    # Test embed_many with deprecated 'texts' parameter emits warning
    with pytest.warns(DeprecationWarning, match="Argument texts is deprecated"):
        embeddings = vectorizer.embed_many(texts=TEST_TEXTS)
    assert isinstance(embeddings, list)
    assert len(embeddings) == len(TEST_TEXTS)



# EXAMPLE: query_vector
# HIDE_START
import json
import warnings
import redis
import numpy as np
from redisvl.index import SearchIndex
from redisvl.query import RangeQuery, VectorQuery
from redisvl.schema import IndexSchema
from sentence_transformers import SentenceTransformer


def embed_text(model, text):
    return np.array(model.encode(text)).astype(np.float32).tobytes()

r = redis.Redis(decode_responses=True)

warnings.filterwarnings("ignore", category=FutureWarning, message=r".*clean_up_tokenization_spaces.*")
model = SentenceTransformer('sentence-transformers/all-MiniLM-L6-v2')

# create index
schema = IndexSchema.from_yaml('data/query_vector_idx.yaml')
index = SearchIndex(schema, r)
index.create(overwrite=True, drop=True)

# load data
with open("data/query_vector.json") as f:
    bicycles = json.load(f)
index.load(bicycles)
# HIDE_END

# STEP_START vector1
query = "Bike for small kids"
query_vector = embed_text(model, query)
print(query_vector[:10]) # >>> b'\x02=c=\x93\x0e\xe0=aC'

vquery = VectorQuery(
    vector=query_vector,
    vector_field_name="description_embeddings",
    num_results=3,
    return_score=True,
    return_fields=["description"]
)
res = index.query(vquery)
print(res) # >>> [{'id': 'bicycle:6b702e8b...', 'vector_distance': '0.399111807346', 'description': 'Kids want...
# REMOVE_START
assert len(res) == 3
# REMOVE_END
# STEP_END

# STEP_START vector2
vquery = RangeQuery(
    vector=query_vector,
    vector_field_name="description_embeddings",
    distance_threshold=0.5,
    return_score=True
).return_fields("description").dialect(2)
res = index.query(vquery)
print(res) # >>> [{'id': 'bicycle:6bcb1bb4...', 'vector_distance': '0.399111807346', 'description': 'Kids want...
# REMOVE_START
assert len(res) == 2
# REMOVE_END
# STEP_END

# REMOVE_START
# destroy index and data
index.delete(drop=True)
# REMOVE_END



"""Integration tests for UNF and NOINDEX field attributes."""

import numpy as np
import pytest

from redisvl.index import SearchIndex
from redisvl.query import FilterQuery, VectorQuery


@pytest.fixture
def sample_data():
    """Create sample data for testing."""
    return [
        {
            "id": "doc1",
            "title": "First Document",
            "content": "This is searchable content",
            "score": 95.5,
            "price": 100,
            "tags": "red,blue",
            "location": "0,0",
            "vector": np.array([0.1, 0.2, 0.3, 0.4], dtype=np.float32).tobytes(),
        },
        {
            "id": "doc2",
            "title": "Second Document",
            "content": "More searchable text here",
            "score": 87.3,
            "price": 200,
            "tags": "green,yellow",
            "location": "1,1",
            "vector": np.array([0.2, 0.3, 0.4, 0.5], dtype=np.float32).tobytes(),
        },
        {
            "id": "doc3",
            "title": "Third Document",
            "content": "Additional content for search",
            "score": 92.1,
            "price": 150,
            "tags": "blue,yellow",
            "location": "2,2",
            "vector": np.array([0.3, 0.4, 0.5, 0.6], dtype=np.float32).tobytes(),
        },
    ]


class TestNoIndexIntegration:
    """Test NOINDEX functionality with real Redis."""

    def test_text_field_with_noindex_not_searchable(self, client, sample_data):
        """Test that TEXT field with NOINDEX cannot be searched."""
        schema = {
            "index": {"name": "test_noindex_text", "prefix": "noindex:"},
            "fields": [
                {"name": "id", "type": "tag"},
                {
                    "name": "title",
                    "type": "text",
                    "attrs": {"no_index": True, "sortable": True},
                },
                {"name": "content", "type": "text"},
            ],
        }

        index = SearchIndex.from_dict(schema, redis_client=client)
        index.create(overwrite=True)
        index.load(sample_data)

        # Should NOT find documents when searching by title (NOINDEX field)
        query = FilterQuery(
            return_fields=["id", "title"],
            filter_expression="@title:(First)",
        )

        # NOINDEX fields return empty results, not an error
        results = index.query(query)
        assert len(results) == 0  # No results because field is not indexed

        # Should find documents when searching by content (indexed field)
        query2 = FilterQuery(
            return_fields=["id", "content"],
            filter_expression="@content:(searchable)",
        )
        results2 = index.query(query2)
        assert len(results2) > 0

        # But title should still be sortable
        query3 = FilterQuery(
            return_fields=["id", "title"],
            filter_expression="*",
            sort_by="title",
        )
        results3 = index.query(query3)
        assert len(results3) == 3
        # Verify sorting worked
        titles = [doc["title"] for doc in results3]
        assert titles == sorted(titles)

        index.delete()

    def test_numeric_field_with_noindex_not_searchable(self, client, sample_data):
        """Test that NUMERIC field with NOINDEX cannot be searched."""
        schema = {
            "index": {"name": "test_noindex_numeric", "prefix": "noindex_num:"},
            "fields": [
                {"name": "id", "type": "tag"},
                {
                    "name": "score",
                    "type": "numeric",
                    "attrs": {"no_index": True, "sortable": True},
                },
                {"name": "price", "type": "numeric"},
            ],
        }

        index = SearchIndex.from_dict(schema, redis_client=client)
        index.create(overwrite=True)
        index.load(sample_data)

        # Should NOT find documents when filtering by score (NOINDEX field)
        query = FilterQuery(
            return_fields=["id", "score"],
            filter_expression="@score:[90 100]",
        )

        # NOINDEX fields return empty results, not an error
        results = index.query(query)
        assert len(results) == 0  # No results because field is not indexed

        # Should find documents when filtering by price (indexed field)
        query2 = FilterQuery(
            return_fields=["id", "price"],
            filter_expression="@price:[100 200]",
        )
        results2 = index.query(query2)
        assert len(results2) >= 2

        # But score should still be sortable
        query3 = FilterQuery(
            return_fields=["id", "score"],
            filter_expression="*",
            sort_by="score",
        )
        results3 = index.query(query3)
        assert len(results3) == 3
        # Verify sorting worked
        scores = [float(doc["score"]) for doc in results3]
        assert scores == sorted(scores)

        index.delete()

    def test_tag_field_with_noindex_not_searchable(self, client, sample_data):
        """Test that TAG field with NOINDEX cannot be searched."""
        schema = {
            "index": {"name": "test_noindex_tag", "prefix": "noindex_tag:"},
            "fields": [
                {"name": "id", "type": "tag"},
                {
                    "name": "tags",
                    "type": "tag",
                    "attrs": {"no_index": True, "sortable": True},
                },
            ],
        }

        index = SearchIndex.from_dict(schema, redis_client=client)
        index.create(overwrite=True)
        index.load(sample_data)

        # Should NOT find documents when filtering by tags (NOINDEX field)
        query = FilterQuery(
            return_fields=["id", "tags"],
            filter_expression="@tags:{blue}",
        )

        # NOINDEX fields return empty results, not an error
        results = index.query(query)
        assert len(results) == 0  # No results because field is not indexed

        # But tags should still be sortable and retrievable
        query2 = FilterQuery(
            return_fields=["id", "tags"],
            filter_expression="*",
            sort_by="tags",
        )
        results2 = index.query(query2)
        assert len(results2) == 3
        # Verify we can retrieve the field values
        assert all("tags" in doc for doc in results2)

        index.delete()

    def test_mixed_index_and_noindex_fields(self, client, sample_data):
        """Test index with mix of indexed and non-indexed fields."""
        schema = {
            "index": {"name": "test_mixed_index", "prefix": "mixed:"},
            "fields": [
                {"name": "id", "type": "tag"},
                {
                    "name": "title",
                    "type": "text",
                    "attrs": {"no_index": True, "sortable": True},
                },
                {"name": "content", "type": "text"},
                {
                    "name": "score",
                    "type": "numeric",
                    "attrs": {"no_index": True, "sortable": True},
                },
                {"name": "price", "type": "numeric"},
                {
                    "name": "vector",
                    "type": "vector",
                    "attrs": {
                        "dims": 4,
                        "distance_metric": "cosine",
                        "algorithm": "flat",
                    },
                },
            ],
        }

        index = SearchIndex.from_dict(schema, redis_client=client)
        index.create(overwrite=True)
        index.load(sample_data)

        # Complex query using only indexed fields
        query = VectorQuery(
            vector=[0.15, 0.25, 0.35, 0.45],
            vector_field_name="vector",
            return_fields=["id", "title", "content", "score", "price"],
            num_results=3,
        )
        results = index.query(query)
        assert len(results) >= 1

        # Verify NOINDEX fields are still returned
        for doc in results:
            assert "title" in doc  # NOINDEX field should still be retrievable
            assert "score" in doc  # NOINDEX field should still be retrievable
            assert "content" in doc
            assert "price" in doc

        index.delete()


class TestUnfIntegration:
    """Test UNF functionality with real Redis."""

    def test_text_field_unf_sortable_unnormalized(self, client):
        """Test that TEXT field with UNF and SORTABLE preserves original case."""
        # Create two indices - one with UNF, one without
        schema_with_unf = {
            "index": {"name": "test_unf_text", "prefix": "unf:"},
            "fields": [
                {"name": "id", "type": "tag"},
                {
                    "name": "title",
                    "type": "text",
                    "attrs": {"unf": True, "sortable": True},
                },
            ],
        }

        schema_without_unf = {
            "index": {"name": "test_no_unf_text", "prefix": "no_unf:"},
            "fields": [
                {"name": "id", "type": "tag"},
                {
                    "name": "title",
                    "type": "text",
                    "attrs": {"sortable": True},  # No UNF
                },
            ],
        }

        # Test data with mixed case
        test_data = [
            {"id": "1", "title": "ZEBRA"},
            {"id": "2", "title": "apple"},
            {"id": "3", "title": "Banana"},
        ]

        # Test with UNF (preserves case for sorting)
        index_unf = SearchIndex.from_dict(schema_with_unf, redis_client=client)
        index_unf.create(overwrite=True)
        index_unf.load(test_data)

        query = FilterQuery(
            return_fields=["id", "title"],
            filter_expression="*",
            sort_by="title",
        )
        results_unf = index_unf.query(query)
        titles_unf = [doc["title"] for doc in results_unf]

        # With UNF, uppercase comes before lowercase in ASCII order
        # Expected order: Banana, ZEBRA, apple (B=66, Z=90, a=97)
        assert titles_unf == ["Banana", "ZEBRA", "apple"]

        # Test without UNF (normalizes to lowercase for sorting)
        index_no_unf = SearchIndex.from_dict(schema_without_unf, redis_client=client)
        index_no_unf.create(overwrite=True)
        index_no_unf.load(test_data)

        query_no_unf = FilterQuery(
            return_fields=["id", "title"],
            filter_expression="*",
            sort_by="title",
        )
        results_no_unf = index_no_unf.query(query_no_unf)
        titles_no_unf = [doc["title"] for doc in results_no_unf]

        # Without UNF, all normalized to lowercase for sorting
        # Expected order: apple, Banana, ZEBRA (alphabetical)
        assert titles_no_unf == ["apple", "Banana", "ZEBRA"]

        index_unf.delete()
        index_no_unf.delete()

    def test_numeric_field_unf_behavior(self, client):
        """Test NUMERIC field UNF behavior - Redis always applies UNF to sortable numeric."""
        schema = {
            "index": {"name": "test_numeric_unf", "prefix": "num_unf:"},
            "fields": [
                {"name": "id", "type": "tag"},
                {
                    "name": "score",
                    "type": "numeric",
                    "attrs": {"sortable": True},  # UNF is implicit for numeric
                },
            ],
        }

        test_data = [
            {"id": "1", "score": 100.5},
            {"id": "2", "score": 50.2},
            {"id": "3", "score": 75.8},
        ]

        index = SearchIndex.from_dict(schema, redis_client=client)
        index.create(overwrite=True)
        index.load(test_data)

        query = FilterQuery(
            return_fields=["id", "score"],
            filter_expression="*",
            sort_by="score",
        )
        results = index.query(query)
        scores = [float(doc["score"]) for doc in results]

        # Numeric sorting should work correctly
        assert scores == [50.2, 75.8, 100.5]

        index.delete()


class TestSchemaRoundtrip:
    """Test that schemas with UNF/NOINDEX can be saved and loaded correctly."""

    def test_schema_persistence_with_new_attributes(self, client, sample_data):
        """Test that index with UNF/NOINDEX can be created and retrieved."""
        schema = {
            "index": {"name": "test_persistence", "prefix": "persist:"},
            "fields": [
                {"name": "id", "type": "tag"},
                {
                    "name": "title",
                    "type": "text",
                    "attrs": {"no_index": True, "sortable": True, "unf": True},
                },
                {
                    "name": "score",
                    "type": "numeric",
                    "attrs": {"no_index": True, "sortable": True},
                },
            ],
        }

        # Create index
        index = SearchIndex.from_dict(schema, redis_client=client)
        index.create(overwrite=True)
        index.load(sample_data)

        # Load index from Redis
        index2 = SearchIndex.from_existing("test_persistence", redis_client=client)

        # Verify fields have correct attributes
        title_field = index2.schema.fields["title"]
        assert title_field.attrs.no_index is True
        assert title_field.attrs.sortable is True
        assert title_field.attrs.unf is True  # Should be preserved for TEXT field

        score_field = index2.schema.fields["score"]
        assert score_field.attrs.no_index is True
        assert score_field.attrs.sortable is True
        # Note: unf for numeric is not preserved as Redis always applies it

        # Verify the index still works
        query = FilterQuery(
            return_fields=["id", "title", "score"],
            filter_expression="*",
            sort_by="title",
        )
        results = index2.query(query)
        assert len(results) == 3

        index.delete()



"""Base cache interface for RedisVL.

This module defines the abstract base cache interface that is implemented by
specific cache types such as LLM caches and embedding caches.
"""

from collections.abc import Mapping
from typing import Any, Dict, Optional, Union

from redis import Redis  # For backwards compatibility in type checking
from redis.cluster import RedisCluster

from redisvl.redis.connection import RedisConnectionFactory
from redisvl.types import AsyncRedisClient, SyncRedisClient, SyncRedisCluster


class BaseCache:
    """Base abstract cache interface for all RedisVL caches.

    This class defines common functionality shared by all cache implementations,
    including TTL management, connection handling, and basic cache operations.
    """

    _redis_client: Optional[SyncRedisClient]
    _async_redis_client: Optional[AsyncRedisClient]

    def __init__(
        self,
        name: str,
        ttl: Optional[int] = None,
        redis_client: Optional[SyncRedisClient] = None,
        async_redis_client: Optional[AsyncRedisClient] = None,
        redis_url: str = "redis://localhost:6379",
        connection_kwargs: Dict[str, Any] = {},
    ):
        """Initialize a base cache.

        Args:
            name (str): The name of the cache.
            ttl (Optional[int], optional): The time-to-live for records cached
                in Redis. Defaults to None.
            redis_client (Optional[SyncRedisClient], optional): A redis client connection instance.
                Defaults to None.
            redis_url (str, optional): The redis url. Defaults to redis://localhost:6379.
            connection_kwargs (Dict[str, Any]): The connection arguments
                for the redis client. Defaults to empty {}.
        """
        self.name = name
        self._ttl: Optional[int] = None
        self.set_ttl(ttl)

        self.redis_kwargs = {
            "redis_client": redis_client,
            "redis_url": redis_url,
            "connection_kwargs": connection_kwargs,
        }

        # Initialize Redis clients
        self._async_redis_client = async_redis_client
        self._redis_client = redis_client

        if redis_client or async_redis_client:
            self._owns_redis_client = False
        else:
            self._owns_redis_client = True

    def _get_prefix(self) -> str:
        """Get the key prefix for Redis keys.

        Returns:
            str: The prefix to use for Redis keys.
        """
        return f"{self.name}:"

    def _make_key(self, entry_id: str) -> str:
        """Generate a full Redis key for the given entry ID.

        Args:
            entry_id (str): The unique entry ID.

        Returns:
            str: The full Redis key including prefix.
        """
        return f"{self._get_prefix()}{entry_id}"

    @property
    def ttl(self) -> Optional[int]:
        """The default TTL, in seconds, for entries in the cache."""
        return self._ttl

    def set_ttl(self, ttl: Optional[int] = None) -> None:
        """Set the default TTL, in seconds, for entries in the cache.

        Args:
            ttl (Optional[int], optional): The optional time-to-live expiration
                for the cache, in seconds.

        Raises:
            ValueError: If the time-to-live value is not an integer.
        """
        if ttl:
            if not isinstance(ttl, int):
                raise ValueError(f"TTL must be an integer value, got {ttl}")
            self._ttl = int(ttl)
        else:
            self._ttl = None

    def _get_redis_client(self) -> SyncRedisClient:
        """Get or create a Redis client.

        Returns:
            SyncRedisClient: A Redis client instance.
        """
        if self._redis_client is None:
            # Create new Redis client
            url = self.redis_kwargs["redis_url"]
            kwargs = self.redis_kwargs["connection_kwargs"]
            self._redis_client = Redis.from_url(url, **kwargs)  # type: ignore
        return self._redis_client

    async def _get_async_redis_client(self) -> AsyncRedisClient:
        """Get or create an async Redis client.

        Returns:
            AsyncRedisClient: An async Redis client instance.
        """
        if not hasattr(self, "_async_redis_client") or self._async_redis_client is None:
            client = self.redis_kwargs.get("redis_client")

            if client and isinstance(client, (Redis, RedisCluster)):
                self._async_redis_client = RedisConnectionFactory.sync_to_async_redis(
                    client
                )
            else:
                url = str(self.redis_kwargs["redis_url"])
                kwargs = self.redis_kwargs.get("connection_kwargs", {})
                if not isinstance(kwargs, Mapping):
                    raise TypeError(
                        "Expected `connection_kwargs` to be a dictionary (e.g. {'decode_responses': True}), "
                        f"but got type: {type(kwargs).__name__}"
                    )
                self._async_redis_client = (
                    RedisConnectionFactory.get_async_redis_connection(url, **kwargs)
                )
        return self._async_redis_client

    def expire(self, key: str, ttl: Optional[int] = None) -> None:
        """Set or refresh the expiration time for a key in the cache.

        Args:
            key (str): The Redis key to set the expiration on.
            ttl (Optional[int], optional): The time-to-live in seconds. If None,
                uses the default TTL configured for this cache instance.
                Defaults to None.

        Note:
            If neither the provided TTL nor the default TTL is set (both are None),
            this method will have no effect.
        """
        _ttl = ttl if ttl is not None else self._ttl
        if _ttl:
            client = self._get_redis_client()
            client.expire(key, _ttl)

    async def aexpire(self, key: str, ttl: Optional[int] = None) -> None:
        """Asynchronously set or refresh the expiration time for a key in the cache.

        Args:
            key (str): The Redis key to set the expiration on.
            ttl (Optional[int], optional): The time-to-live in seconds. If None,
                uses the default TTL configured for this cache instance.
                Defaults to None.

        Note:
            If neither the provided TTL nor the default TTL is set (both are None),
            this method will have no effect.
        """
        _ttl = ttl if ttl is not None else self._ttl
        if _ttl:
            client = await self._get_async_redis_client()
            await client.expire(key, _ttl)

    def clear(self) -> None:
        """Clear the cache of all keys."""
        client = self._get_redis_client()
        prefix = self._get_prefix()

        # Scan for all keys with our prefix
        cursor = 0  # Start with cursor 0
        while True:
            cursor_int, keys = client.scan(cursor=cursor, match=f"{prefix}*", count=100)  # type: ignore
            if keys:
                client.delete(*keys)
            if cursor_int == 0:  # Redis returns 0 when scan is complete
                break
            # Cluster returns a dict of cursor values. We need to stop if these all
            # come back as 0.
            elif isinstance(cursor_int, Mapping):
                cursor_values = list(cursor_int.values())
                if all(v == 0 for v in cursor_values):
                    break
            else:
                cursor = cursor_int  # Update cursor for next iteration

    async def aclear(self) -> None:
        """Async clear the cache of all keys."""
        client = await self._get_async_redis_client()
        prefix = self._get_prefix()

        # Scan for all keys with our prefix
        cursor = 0  # Start with cursor 0
        while True:
            cursor_int, keys = await client.scan(
                cursor=cursor, match=f"{prefix}*", count=100
            )  # type: ignore
            if keys:
                await client.delete(*keys)
            if cursor_int == 0:  # Redis returns 0 when scan is complete
                break
            # Cluster returns a dict of cursor values. We need to stop if these all
            # come back as 0.
            elif isinstance(cursor_int, Mapping):
                cursor_values = list(cursor_int.values())
                if all(v == 0 for v in cursor_values):
                    break
            else:
                cursor = cursor_int  # Update cursor for next iteration

    def disconnect(self) -> None:
        """Disconnect from Redis."""
        if self._owns_redis_client is False:
            return

        if self._redis_client:
            self._redis_client.close()
            self._redis_client = None
            # Async clients don't have a sync close method, so we just
            # zero them out to allow garbage collection.
            self._async_redis_client = None

    async def adisconnect(self) -> None:
        """Async disconnect from Redis."""
        if self._owns_redis_client is False:
            return

        if self._redis_client:
            self._redis_client.close()
            self._redis_client = None

        if hasattr(self, "_async_redis_client") and self._async_redis_client:
            # Use proper async close method
            await self._async_redis_client.aclose()
            self._async_redis_client = None
